#include "framecpp/config.h"

#include "general/mutexlock.hh"
#include "general/Singleton.hh"
#include "general/gpstime.hh"

#include "framecpp/Common/Description.hh"
#include "framecpp/Common/FrameStream.hh"
#include "framecpp/Common/FrameSpec.tcc"

#include "framecpp/Version8/FrProcData.hh"
#include "framecpp/Version8/FrSE.hh"
#include "framecpp/Version8/FrSH.hh"

#include "framecpp/Version8/PTR_STRUCT.hh"

using std::ceil;
using std::floor;

using namespace FrameCPP::Version_8;
using FrameCPP::Common::Description;
using FrameCPP::Common::FrameSpec;
using FrameCPP::Common::IStream;
using FrameCPP::Common::OStream;

//=======================================================================
// Static
//=======================================================================

static const FrameSpec::Info::frame_object_types s_object_id
= FrameSpec::Info::FSI_FR_PROC_DATA;

//=======================================================================
// Anonymous
//=======================================================================
namespace {
  class IDTypeToStringMapping
  {
    SINGLETON_TS_DECL( IDTypeToStringMapping );
  public:
    const std::string& GetString( INT_2U Id ) const;

  private:
    std::vector< std::string >	m_types;
  };

  class IDSubTypeToStringMappingFrequency
  {
    SINGLETON_TS_DECL( IDSubTypeToStringMappingFrequency );
  public:
    const std::string& GetString( INT_2U Id ) const;

  private:
    std::vector< std::string >	m_types;
  };

} // namespace - anonymous


//=======================================================================
// FrProcData::fr_proc_data_ref_type
//=======================================================================

FrProcData::fr_proc_data_ref_type::
fr_proc_data_ref_type( )
{
}

FrProcData::fr_proc_data_ref_type::
fr_proc_data_ref_type( Common::IStream& Stream )
{
  Stream >> data
	 >> aux
	 >> table
	 >> history
    ;
}

static INT_2U MAX_REF = 4;

//=======================================================================
// FrProcData
//=======================================================================
FrProcData::
FrProcData( )
  : object_type( s_object_id, StructDescription( ) ),
    m_synced_with_vector( false )
{
}

FrProcData::
FrProcData( const FrProcData& ProcData )
  : object_type( s_object_id, StructDescription( ) ),
    FrProcDataData( ProcData ),
    m_refs( ProcData.m_refs ),
    m_synced_with_vector( false )
{
}

FrProcData::
FrProcData( const std::string& Name,
	    const std::string& Comment,
	    INT_2U Type,
	    INT_2U SubType,
	    REAL_8 TimeOffset,
	    REAL_8 TRange,
	    REAL_8 FShift,
	    REAL_4 Phase,
	    REAL_8 FRange,
	    REAL_8 BW )
  : object_type( s_object_id, StructDescription( ) ),
    m_synced_with_vector( false )
{
  m_data.name = Name;
  m_data.comment = Comment;
  m_data.type = Type;
  m_data.subType = SubType;
  m_data.timeOffset = TimeOffset;
  m_data.tRange = TRange;
  m_data.fShift = FShift;
  m_data.phase = Phase;
  m_data.fRange = FRange;
  m_data.BW = BW;
}

FrProcData::
FrProcData( const std::string& name,
	    REAL_8 sampleRate, const GPSTime& timeOffset,
	    REAL_8 fShift, REAL_4 phase )
  : object_type( s_object_id, StructDescription( ) ),
    m_synced_with_vector( false )

{
  m_data.name = name;
  m_data.comment = std::string( "" );
  m_data.type = 0;
  m_data.subType = 0;
  m_data.timeOffset = timeOffset - GPSTime( 0, 0 );
  m_data.tRange = 0;
  m_data.fShift = fShift;
  m_data.phase = phase;
  m_data.fRange = 0;
  m_data.BW = 0;
}

FrProcData::
FrProcData( Previous::FrProcData& Source, IStream* Stream )
  : object_type( s_object_id, StructDescription( ) ),
    FrProcDataData( Source ),
    m_synced_with_vector( false )
{
  if ( Stream )
  {
    //-------------------------------------------------------------------
    // Modify references
    //-------------------------------------------------------------------
    Stream->ReplaceRef( RefData( ), &(Source.RefData( )), MAX_REF );
    Stream->ReplaceRef( RefAux( ), &(Source.RefAux( )), MAX_REF );
    Stream->ReplaceRef( RefTable( ), &(Source.RefTable( )), MAX_REF );
    Stream->ReplaceRef( RefHistory( ), &(Source.RefHistory( )), MAX_REF );
  }
}

FrProcData::
FrProcData( IStream& Stream )
  : object_type( s_object_id, StructDescription( ) ),
    FrProcDataData( Stream ),
    m_refs( Stream ),
    m_synced_with_vector( false )
{
  Stream.Next( this );
}

FrProcData::
~FrProcData( )
{
}

const std::string& FrProcData::
IDTypeToString( INT_2U Type )
{
  return IDTypeToStringMapping::Instance( ).GetString( Type );
}

const char* const FrProcData::
ObjectStructName( ) const
{
  return StructName( );
}

const std::string& FrProcData::
GetNameSlow( ) const
{
  return GetName( );
}

const Description* FrProcData::
StructDescription( )
{
  static Description ret;

  if ( ret.size( ) == 0 )
  {
    ret( FrSH( FrProcData::StructName(), s_object_id,
	       "Post-processed Data Structure" ) );

    ret( FrSE( "name", "STRING",
	       "" ) );
    ret( FrSE( "comment", "STRING",
	       "" ) );
    ret( FrSE( "type", "INT_2U",
	       "" ) );
    ret( FrSE( "subType", "INT_2U",
	       "" ) );
    ret( FrSE( "timeOffset", "REAL_8",
	       "" ) );
    ret( FrSE( "tRange", "REAL_8",
	       "" ) );
    ret( FrSE( "fShift", "REAL_8",
	       "" ) );
    ret( FrSE( "phase", "REAL_4",
	       "" ) );
    ret( FrSE( "fRange", "REAL_8",
	       "" ) );
    ret( FrSE( "BW", "REAL_8",
	       "" ) );
    ret( FrSE( "nAuxParam", "INT_2U",
	       "" ) );
    ret( FrSE( "auxParam", "REAL_8[nAuxParam]",
	       "" ) );
    ret( FrSE( "auxParamNames", "STRING[nAuxParam]",
	       "" ) );

    ret( FrSE( "data",	PTR_STRUCT::Desc( FrVect::StructName( ) ) ) );
    ret( FrSE( "aux", PTR_STRUCT::Desc( FrVect::StructName( ) ) ) );
    ret( FrSE( "table", PTR_STRUCT::Desc( FrTable::StructName( ) ) ) );
    ret( FrSE( "history", PTR_STRUCT::Desc( FrHistory::StructName( ) ) ) );

    ret( FrSE( "next", PTR_STRUCT::Desc( FrProcData::StructName( ) ) ) );

    ret( FrSE( "chkSum", CheckSumDataClass( ), CheckSumDataComment( ) ) );
  }

  return &ret;
}

const std::string& FrProcData::
IDSubTypeToString( INT_2U Type, INT_2U SubType )
{
  static const std::string	empty( "" );
  switch ( Type )
  {
  case FREQUENCY_SERIES:
  return IDSubTypeToStringMappingFrequency::Instance( ).GetString( Type );
    break;
  default:
    return empty;
    break;
  }
}

FrProcData& FrProcData::
Merge( const FrProcData& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += StructName( );

  throw std::domain_error( msg );
  return *this;
}

std::auto_ptr< FrProcData > FrProcData::
SubFrProcData( REAL_8 Offset, REAL_8 Dt ) const
{
  std::auto_ptr< FrProcData >	retval;

  //---------------------------------------------------------------------
  // Process the types that are understood
  //---------------------------------------------------------------------
  switch( GetType( ) )
  {
  case TIME_SERIES:
    retval = sub_time_series( Offset, Dt );
    break;
  default:
    {
      //-----------------------------------------------------------------
      // This type is not supported so throw an exception
      //-----------------------------------------------------------------
      std::ostringstream	msg;
      msg << "Currently cannot take a subsection of type: "
	  << IDTypeToString( GetType( ) );
      const std::string& subtype( IDSubTypeToString( GetType( ),
						     GetSubType( ) ) );
      if ( subtype.length( ) > 0 )
      {
	msg << " subType: " << subtype;
      }
      throw std::runtime_error( msg.str( ) );
    }
    break;
  }

  return retval;
}

bool FrProcData::
operator==( const Common::FrameSpec::Object& Obj ) const
{
  return ( *this == Obj );
}

void FrProcData::
copy_core( const FrProcData& Source )
{
  static_cast< FrProcDataData& >( *this ) = Source;
  m_refs.copy_core( Source.m_refs );
}

std::auto_ptr< FrameSpec::Object > FrProcData::
demote( INT_2U Target, IStream* Stream ) const
{
  if ( Target >= DATA_FORMAT_VERSION )
  {
    return std::auto_ptr< FrameSpec::Object >
      ( const_cast< FrProcData* >( this ) );
  }
  try
  {
    //-------------------------------------------------------------------
    // Copy non-reference information
    //-------------------------------------------------------------------
    // Do actual down conversion
    /// \todo See if the actual sample rate can be computed
    Previous::GPSTime t;
    t += GetTimeOffset( );
    std::auto_ptr< Previous::FrProcData >
      retval( new Previous::FrProcData( GetName( ),
					GetComment( ),
					GetType( ),
					GetSubType( ),
					GetTimeOffset( ),
					GetTRange( ),
					GetFShift( ),
					GetPhase( ),
					GetFRange( ),
					GetBW( )
					) )
      ;
    if ( Stream )
    {
      //-----------------------------------------------------------------
      // Modify references
      //-----------------------------------------------------------------
      Stream->ReplaceRef( retval->RefAux( ), &(RefAux( )), MAX_REF );
      Stream->ReplaceRef( retval->RefData( ), &(RefData( )), MAX_REF );
      Stream->ReplaceRef( retval->RefTable( ), &(RefTable( )), MAX_REF );
      Stream->ReplaceRef( retval->RefHistory( ), &(RefHistory( )), MAX_REF );
    }
    //-------------------------------------------------------------------
    // Return demoted object
    //-------------------------------------------------------------------
    return std::auto_ptr< FrameSpec::Object >( retval.release( ) );
  }
  catch( ... )
  {
  }
  throw
    Unimplemented( "Object* FrProcData::demote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}

std::auto_ptr< FrameSpec::Object > FrProcData::
promote( INT_2U Target, Object* Obj, IStream* Stream ) const
{
  return Promote( Target, Obj, Stream );
}

FrameCPP::cmn_streamsize_type FrProcData::
pBytes( const Common::StreamBase& Stream ) const
{
  return
    m_data.name.Bytes( )
    + m_data.comment.Bytes( )
    + sizeof( m_data.type )
    + sizeof( m_data.subType )
    + sizeof( m_data.timeOffset )
    + sizeof( m_data.tRange )
    + sizeof( m_data.fShift )
    + sizeof( m_data.phase )
    + sizeof( m_data.fRange )
    + sizeof( m_data.BW )
    + m_data.auxParams.Bytes( Stream )
    + Stream.PtrStructBytes( ) // data
    + Stream.PtrStructBytes( ) // aux
    + Stream.PtrStructBytes( ) // table
    + Stream.PtrStructBytes( ) // history
    + Stream.PtrStructBytes( ) // next
    ;
}

FrProcData* FrProcData::
pCreate( IStream& Stream ) const
{
  return new FrProcData( Stream );
}

void FrProcData::
pWrite( OStream& Stream ) const
{
  m_data( Stream );
  Stream << m_refs.data
	 << m_refs.aux
	 << m_refs.table
	 << m_refs.history
      ;
  WriteNext( Stream );
}

void FrProcData::
sync_with_vector( )
{
  if ( ( m_synced_with_vector == false ) &&
       ( RefData( ).size( ) > 0 ) )
  {
    switch( GetType( ) )
    {
    case TIME_SERIES:
      m_data.tRange = ( RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) *
			RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
      break;
    case FREQUENCY_SERIES:
      m_data.fRange = ( RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) *
			RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
      break;
    }
    m_synced_with_vector = true;
  }
}

// Routine to take a subsection when the data type is TimeSeries data
std::auto_ptr< FrProcData > FrProcData::
sub_time_series( REAL_8 Offset, REAL_8 Dt ) const
{
  //---------------------------------------------------------------------
  // Sanity checks
  //---------------------------------------------------------------------
  // Verify the existance of only a single FrVect component
  //---------------------------------------------------------------------
  if ( m_refs.data.size( ) != 1 )
  {
    std::ostringstream	msg;

    msg << "METHOD: frameAPI::FrProcData::SubFrProcData: "
	<< "data vectors of length " << m_refs.data.size( )
	<< " currently are not supported";
    throw std::logic_error( msg.str( ) );
  }
  INT_4U	num_elements = m_refs.data[0]->GetDim( 0 ).GetNx( );
  //---------------------------------------------------------------------
  //:TODO: Ensure the dimension of the FrVect is 1
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  //:TODO: Ensure the requested range is valid
  //---------------------------------------------------------------------
  REAL_8	sample_rate = 1.0 / m_refs.data[ 0 ]->GetDim( 0 ).GetDx( );
  if ( Offset > ( m_data.timeOffset +
		  ( num_elements / sample_rate ) ) )
  {
    std::ostringstream msg;
    msg << "METHOD: frameAPI::FrProcData::SubFrProcData: "
	<< "The value for the parameter Offset (" << Offset
	<< ") is out of range ([0-"
	<< ( m_data.timeOffset + ( num_elements / sample_rate ) )
	<< "))";
    throw std::range_error( msg.str( ) );
  }
  //---------------------------------------------------------------------
  // Start moving of data
  //---------------------------------------------------------------------
  std::auto_ptr< FrProcData >	retval( new FrProcData );
  // Copy in the header information
  retval->copy_core( *this );

  //:TODO: Get the slice of FrVect
  REAL_8	startSampleReal =
    ( Offset * sample_rate ) -
    ( retval->m_data.timeOffset * sample_rate );

  INT_4U	endSample =
    INT_4U( ceil ( startSampleReal +
		   ( Dt * sample_rate ) ) );
  INT_4U	startSample = ( startSampleReal > REAL_8( 0 ) )
    ? INT_4U( floor( startSampleReal ) )
    : INT_4U( 0 );

  retval->m_refs.data.append( this->m_refs.data[ 0 ]->SubFrVect( startSample, endSample ).release( ),
			      false, /* allocate */
			      true /* own */ );
  //---------------------------------------------------------------------
  // Shift the offset
  //---------------------------------------------------------------------
  if ( startSampleReal > 0 )
  {
    // There is no gap
    retval->m_data.timeOffset = REAL_8( 0 );
  }
  else
  {
    retval->m_data.timeOffset += Offset;
  }
  //---------------------------------------------------------------------
  // :TODO: Recalculate tRange
  //---------------------------------------------------------------------
  retval->SetTRange( retval->RefData( )[ 0 ]->GetDim( 0 ).GetNx( ) * 
		     retval->RefData( )[ 0 ]->GetDim( 0 ).GetDx( ) );
  //---------------------------------------------------------------------
  // Return the new FrProcData
  //---------------------------------------------------------------------
  return retval;
}

//=======================================================================
// Anonymous
//=======================================================================
namespace
{
  //---------------------------------------------------------------------
  // Initialization of singleton
  //---------------------------------------------------------------------
  SINGLETON_TS_INST( IDTypeToStringMapping );
  //---------------------------------------------------------------------
  // Default constructor
  //---------------------------------------------------------------------
  IDTypeToStringMapping::
  IDTypeToStringMapping( )
  {
    m_types.push_back( "Unknown/user defined" );
    m_types.push_back( "Time Series" );
    m_types.push_back( "Frequency Series" );
    m_types.push_back( "Other 1D Series data" );
    m_types.push_back( "Time-Frequency" );
    m_types.push_back( "Wavelets" );
    m_types.push_back( "Multi-dimensional" );
  }

  const std::string& IDTypeToStringMapping::
  GetString( INT_2U Type ) const
  {
    if ( Type >= m_types.size( ) )
    {
      std::ostringstream	msg;
      msg << "Cannot map FrProcData type value of " << Type
	  << " to string value because it is out of range";
      throw std::range_error( msg.str( ) );
    }
    return m_types[ Type ];
  }

  //---------------------------------------------------------------------
  // Initialization of singleton
  //---------------------------------------------------------------------
  SINGLETON_TS_INST( IDSubTypeToStringMappingFrequency );
  //---------------------------------------------------------------------
  // Default constructor
  //---------------------------------------------------------------------
  IDSubTypeToStringMappingFrequency::
  IDSubTypeToStringMappingFrequency( )
  {
    m_types.push_back( "Unknown/user defined" );
    m_types.push_back( "Amplitude Spectral Density" );
    m_types.push_back( "Power Spectral Density" );
    m_types.push_back( "Cross Spectral Density" );
    m_types.push_back( "Coherence" );
    m_types.push_back( "Transfer Function" );
  }

  const std::string& IDSubTypeToStringMappingFrequency::
  GetString( INT_2U Type ) const
  {
    if ( Type >= m_types.size( ) )
    {
      std::ostringstream	msg;
      msg << "Cannot map FrProcData type value of " << Type
	  << " to string value because it is out of range";
      throw std::range_error( msg.str( ) );
    }
    return m_types[ Type ];
  }
} // namespace - anonymous

void FrProcData::
TOCQuery( int InfoClass, ... ) const
{
  using Common::TOCInfo;

  va_list	vl;
  va_start( vl, InfoClass );
  while ( InfoClass != TOCInfo::IC_EOQ )
  {
    int data_type = va_arg( vl, int );
    switch( data_type )
    {
    case TOCInfo::DT_STRING_2:
      {
	STRING* data = va_arg( vl, STRING* );
	switch( InfoClass )
	{
	case TOCInfo::IC_NAME:
	  *data = GetName( );
	  break;
	default:
	  goto cleanup;
	  break;
	}
      }
      break;
    default:
      // Stop processing
      goto cleanup;
    }
    InfoClass = va_arg( vl, int );
  }
 cleanup:
  va_end( vl );
}

