#ifndef GENERAL__READ_WRITE_LOCK_HH
#define	GENERAL__READ_WRITE_LOCK_HH

#include "general/config.h"

#include <stdexcept>
#include <string>

#include "general/Thread.hh"

namespace General
{
  class ReadWriteLock
  {
  public:
    enum mode_type
    {
      READ,
      WRITE
    };

    class BusyError
      : public std::runtime_error
    {
    public:
      BusyError( );
    private:
      std::string msg( );
    };

    class TimedoutError
      : public std::runtime_error
    {
    public:
      TimedoutError( );
    private:
      std::string msg( );
    };

    class Sync
    {
    public:
#ifdef HAVE_PTHREAD_RW_LOCK
      typedef pthread_rwlock_t	lock_type;
#else /* HAVE_PTHREAD_RW_LOCK */
      typedef pthread_mutex_t	lock_type;
#endif /* HAVE_PTHREAD_RW_LOCK */
      static lock_type	INITIALIZER;

      Sync( );
      Sync( lock_type& Source );
      Sync( lock_type* Source );

      void Lock( );
      void Unlock( );
      void ReadLock( );
      void ReadUnlock( );

    private:
      lock_type		m_lock;
      lock_type*	m_lock_ptr;

      Sync( const Sync& Source );
      Sync& operator=( const Sync& Source );
    };

    typedef Sync::lock_type	lock_type;

    ReadWriteLock( lock_type& Lock, mode_type Mode,
		   const char* Filename = "",
		   int Linenum = 0 );

    ReadWriteLock( lock_type& Lock, mode_type Mode,
		   bool TryLock,
		   const char* Filename = "",
		   int Linenum = 0 );

    ReadWriteLock( lock_type& Lock, mode_type Mode, int Timeout,
		   const char* Filename = "",
		   int Linenum = 0 );

    virtual ~ReadWriteLock( );

    static void Destroy( lock_type& Lock );
    static void Initialize( lock_type& Lock );
    void Release( const char* Filename = "",
		  int Linenum = 0 );


    static int Interval( int Value );
    static int Timeout( int Value );

    //-------------------------------------------------------------------
    /// \brief Handler for thead cancelation
    ///
    /// \param[in] VLock
    ///     The VLock is a pointer to a ReadWriteLock object that holds
    ///     a lock.
    //-------------------------------------------------------------------
    static void ThreadCancel( void* VLock );

  private:
    ReadWriteLock( const ReadWriteLock& Source );
    const ReadWriteLock& operator=( const ReadWriteLock& Source );

    static void thread_cleanup( void* VLock );

    lock_type*	m_lock;
    mode_type	m_mode;
    int		m_old_cancel_state;
  }; // class ReadWriteLock

} // namespace General

#endif	/* GENERAL__READ_WRITE_LOCK_HH */
